# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

from astropy.io import fits
import numpy as np

class ObjectMask:
    """
    ObjectMask is a class to generate an object mask, which is a 2D image of boolean for True = object.
    - nx,ny is the shape of the image
    - tx,ty is the trace
    - halfdyup,halfdylow is the aperture width from the trace (i.e., the total width is halfdylow + 1 + halfdyup)
    - Note: given (txi,tyi), image[tyi-halfdylow:tyi+1+halfdyup,txi] = True    
    Use method compute() to generate the object mask, which is accessible through self.mask.
    Use method save(container) to save the mask image as ./savefolder/saveprefix_mask.fits given container.
    - container is an object of hstgrism.container.Container class
    Note: if cutout.fits and bbcorner.csv was created from Cutout2D, together with trace.csv from earlier process, a simple way to create an object mask is:
    - Set ny,nx = gdata.shape where gdata is the image from cutout.fits
    - tx = trace.xh + trace.xyref[0] - bbcorner.bb0x[0] where trace from trace.csv and bbcorner from bbcorner.csv
    - ty is similar to tx
    - halfdyup, halfdylow are arbitrary
    """
    def __init__(self,nx,ny,tx,ty,halfdyup,halfdylow):
        self.nx = nx
        self.ny = ny
        self.tx = tx
        self.ty = ty
        self.halfdyup = halfdyup
        self.halfdylow = halfdylow
        self.mask = None
    def compute(self):
        tmp = np.arange(self.nx*self.ny).reshape(self.ny,self.nx)
        tmp = np.full_like(tmp,False,dtype=bool)
        for i,ii in enumerate(self.tx):
            xx,yy = self.tx[i],self.ty[i]
            if int(xx) < 0 or int(xx) > self.nx - 1:
                continue
            ymin = max(0,int(yy-self.halfdylow))
            ymax = min(self.ny,int(yy+1+self.halfdyup))
            tmp[ymin:ymax,int(xx)] = True        
        self.mask = tmp.copy()
    def save(self,container=None):
        if container is None:
            raise ValueError('container must be specified to save')
        savefolder = container.data['savefolder']
        saveprefix = container.data['saveprefix']
        phdu = fits.PrimaryHDU()
        imhdu = fits.ImageHDU()
        hdul = fits.HDUList([phdu,imhdu])
        hdul[1].data = self.mask.astype(int)
        string = './{0}/{1}_mask.fits'.format(savefolder,saveprefix)
        hdul.writeto(string,overwrite=True)
        print('Save {0}'.format(string))
    


# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

import os,glob,warnings

class Container:
    """
    Container is a class facilitating how the output files would be named, and packaged in a folder.
    Name convention of HSTGRISM files is ./savefolder/saveprefix_suffix.ext.
    suffix and ext would be chosen internally in the pipeline. If it is a graphic file such as plots, plotformat determines its ext.
    overwrite, if set True, an existing folder with the same name would be removed at the beginning.
    """
    def __init__(self,saveprefix,savefolder,plotformat,overwrite):
        self.data = {'saveprefix':saveprefix,
                     'savefolder':savefolder,
                     'plotformat':plotformat,
                     'overwrite':overwrite
                    }
        self._prep_container()
    def _prep_container(self):
        tmp = glob.glob('*')
        if self.data['savefolder'] in tmp:
            sentinel = True
        else:
            sentinel = False
        if not sentinel:
            os.mkdir(self.data['savefolder'])
        else:
            if not self.data['overwrite']:
                string = 'Folder {0} already exists. To create a fresh folder, set overwrite = True'.format(self.data['savefolder'])
                warnings.warn(string)
            else:
                os.system('rm -r ./{0}'.format(self.data['savefolder']))
                os.mkdir(self.data['savefolder'])



# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

import numpy as np
from scipy.interpolate import interp2d
import copy

class GrismApCorr:
    """
    GrismApCorr is a class containing tables for aperture correction (i.e., apcorr = f(wavelength, apsize)) in aXe reduction. These tables are from ISRs. Interpolaton model is also prepared.
    - Available instrument can be found by GrismApCorr().available_instrument.
    - apsize is an array with the same length as wave, and must be diameter in arcsec. This can be computed as apsize = (halfdyup + 1 + halfdylow) * scale where (halfdyup,halfdylow) can be found in tbox.csv (if generated from earlier process), and scale can be found from GrismApCorr().table[instrument]['scale'].
    - Note: halfdyup must be equal to halfdylow according to the ISRs.
    - wave is an array of wavelengths, and must be in Angstrom.
    Use method compute() after instantiating the GrismApCorr object correctly to return the aperture correction values, which is accessible through self.data['apcorr'].
    - Note: compute(fill_value='median') is the only implementation in this version. This will replace any out-of-bound values with median value calculated from in-bound values.
    ##########
    x = GrismApCorr()
    x.table = show apcorr tables with keys as intrument codes. Use x.instrument to check which instrument is available.
    x.instrument = show available instrument codes
    x.table[instrument]['model'](wave,apsize) = calculate apcorr given a model corresponding to its instrument, wavelength, and aperture size.
    ##########
    Note: to avoid crash, and to support wider wavelength, we manually added min and max wavelengths to the tables.
    - for HST-WFC3-IR-G102: we added 7000. and 12000.
    - for HST-WFC3-IR-G141: we added 11000. and 17000.
    - for HST-WFC3-UVIS1-G280: we added 2000. and 8000.
    - apcorr values for these additional columns were set to be the nearest available values.
    - This change would simply make the interpolation of apcorr outside the original bounds to be nearest neighbour.
    """    
    def __init__(self,instrument=None,apsize=None,wave=None,aptype='diameter',apunit='arcsec',waveunit='A'):
        self.table = self._get_table()
        self.available_instrument = self.table.keys()
        self.data = {'instrument':None,
                     'apsize':None,
                     'aptype':aptype,
                     'apunit':apunit,
                     'wave':None,
                     'waveunit':waveunit,
                     'apcorr':None
                    }
        if instrument is not None:
            self.data['instrument'] = instrument
        if apsize is not None:
            self.data['apsize'] = apsize
        if wave is not None:
            self.data['wave'] = wave
    def compute(self,fill_value='median'):
        if fill_value == 'median':
            fill = np.nan
        instrument = self.data['instrument']
        apsize = self.data['apsize']
        aptype = self.data['aptype']
        apunit = self.data['apunit']
        wave = self.data['wave']
        waveunit = self.data['waveunit']
        table = self.table[instrument]
        # check aptype
        if aptype != table['type']:
            raise ValueError('aptype not equal {0}'.format(table['type']))
        # check apunit
        if apunit != table['apunit']:
            if table['apunit'] == 'arcsec' and table['scaleunit'] == 'arcsec/pix' and apunit == 'pix':
                apsize = apsize * table['scale']
                apunit = 'arcsec'
                self.data['apsize'] = apsize
                self.data['apunit'] = apunit
                print('Scale apsize in {0} with {1} {2}'.format(apunit,table['scale'],table['scaleunit']))
            else:
                raise ValueError('apunit not equal {0} or pix'.format(table['apunit']))
        # check waveunit
        if waveunit != table['waveunit']:
            raise ValueError('waveunit not equal {0}'.format(table['waveunit']))
        model = self._model(table['wave'],table['apsize'],table['value'],fill)
        apcorr = model(wave,apsize).diagonal()
        if fill_value == 'median':
            tmp = apcorr.copy()
            m_good = np.argwhere(np.isfinite(tmp)).flatten()
            m_bad = np.argwhere(~np.isfinite(tmp)).flatten()
            median = np.median(tmp[m_good])
            tmp[m_bad] = median
            apcorr = tmp.copy()
            print('Replace bad apcorr values with median = {0:.3f}'.format(median))
        self.data['apcorr'] = apcorr
    def _model(self,wave,apsize,value,fill_value):
        model = interp2d(wave,apsize,value,kind='linear',copy=True,
                         bounds_error=False,fill_value=fill_value
                        )
        return model  
    def _get_table(self):
        ##########
        # Tables from ISRs
        TABLE = {'HST-WFC3-IR-G102': 
                 {'ref': 'ISR WFC3-2011-05'
                  ,'filter': 'G102'
                  ,'scale': 0.13
                  ,'scaleunit': 'arcsec/pix'
                  ,'type': 'diameter'
                  ,'row': 'apsize'
                  ,'col': 'wave'
                  ,'apunit': 'arcsec'
                  ,'apsize': np.array((0.128,0.385,0.641
                                       ,0.898,1.154,1.411
                                       ,1.667,1.924,3.719
                                       ,7.567,12.954,25.779
                                      ))
                  ,'waveunit': 'A'
                  ,'wave': np.array((7000.,8850.,9350.,9850.,10350.,10850.,11350.,12000.))
                  ,'value' : np.array(((0.459,0.459,0.391,0.414,0.464,0.416,0.369,0.369)
                                      ,(0.825,0.825,0.809,0.808,0.811,0.794,0.792,0.792)
                                      ,(0.890,0.890,0.889,0.887,0.880,0.875,0.888,0.888)
                                      ,(0.920,0.920,0.917,0.916,0.909,0.904,0.916,0.916)
                                      ,(0.939,0.939,0.937,0.936,0.930,0.925,0.936,0.936)
                                      ,(0.952,0.952,0.950,0.950,0.943,0.940,0.949,0.949)
                                      ,(0.962,0.962,0.961,0.961,0.954,0.951,0.958,0.958)
                                      ,(0.969,0.969,0.968,0.969,0.962,0.959,0.965,0.965)
                                      ,(0.985,0.985,0.984,0.986,0.982,0.980,0.983,0.983)
                                      ,(0.995,0.995,0.995,0.996,0.991,0.990,0.992,0.992)
                                      ,(0.999,0.999,0.999,0.999,0.997,0.996,0.995,0.995)
                                      ,(1.000,1.000,1.000,1.000,1.000,1.000,1.000,1.000)
                                     ))
                  ,'model': None
                 }
                 ,'HST-WFC3-IR-G141':
                 {'ref': 'ISR WFC3-2011-05'
                  ,'filter': 'G141'
                  ,'scale': 0.13
                  ,'scaleunit': 'arcsec/pix'
                  ,'type': 'diameter'
                  ,'row': 'apsize'
                  ,'col': 'wave'
                  ,'apunit': 'arcsec'
                  ,'apsize': np.array((0.128,0.385,0.641
                                       ,0.898,1.154,1.411
                                       ,1.667,1.924,3.719
                                       ,7.567,12.954,25.779
                                      ))   
                  ,'waveunit': 'A'
                  ,'wave': np.array((11000.,11300.,12300.,13300.,14300.,15300.,16300.,17000.))
                  ,'value': np.array(((0.442,0.442,0.444,0.395,0.344,0.342,0.376,0.376)
                                     ,(0.805,0.805,0.792,0.764,0.747,0.732,0.732,0.732)
                                     ,(0.866,0.866,0.877,0.865,0.863,0.850,0.859,0.859)
                                     ,(0.912,0.912,0.901,0.893,0.894,0.884,0.898,0.898)
                                     ,(0.933,0.933,0.924,0.914,0.913,0.903,0.913,0.913)
                                     ,(0.947,0.947,0.940,0.931,0.932,0.921,0.932,0.932)
                                     ,(0.958,0.958,0.950,0.942,0.944,0.934,0.945,0.945)
                                     ,(0.966,0.966,0.959,0.951,0.953,0.944,0.954,0.954)
                                     ,(0.985,0.985,0.984,0.981,0.985,0.980,0.985,0.985)
                                     ,(0.993,0.993,0.995,0.992,0.997,0.992,0.996,0.996)
                                     ,(0.996,0.996,0.998,0.997,1.000,0.997,1.000,1.000)
                                     ,(1.000,1.000,1.000,1.000,1.000,1.000,1.000,1.000)
                                    ))
                  ,'model': None
                 }
                 ,'HST-WFC3-UVIS1-G280':
                 {'ref': 'ISR WFC3-2009-01'
                  ,'filter': 'G280'
                  ,'scale': 0.04
                  ,'scaleunit': 'arcsec/pix'
                  ,'type': 'diameter'
                  ,'row': 'apsize'
                  ,'col': 'wave'
                  ,'apunit': 'arcsec'
                  ,'apsize': np.array((0.04,0.12,0.20,
                                       0.28,0.36,0.44,
                                       0.52,0.60,1.16,
                                       2.36,4.04,8.04
                                      ))   
                  ,'waveunit': 'A'
                  ,'wave': np.array((2000.,2275.,2825.,3375.,3925.,4475.,5025.,8000.))
                  ,'value': np.array(((0.312,0.312,0.312,0.276,0.258,0.261,0.241,0.241)
                                     ,(0.720,0.720,0.706,0.676,0.642,0.600,0.589,0.589)
                                     ,(0.856,0.856,0.848,0.839,0.815,0.767,0.769,0.769)
                                     ,(0.903,0.903,0.899,0.897,0.882,0.840,0.855,0.855)
                                     ,(0.926,0.926,0.926,0.926,0.913,0.875,0.895,0.895)
                                     ,(0.939,0.939,0.942,0.944,0.933,0.898,0.916,0.916)
                                     ,(0.947,0.947,0.951,0.956,0.947,0.915,0.931,0.931)
                                     ,(0.953,0.953,0.957,0.963,0.956,0.928,0.944,0.944)
                                     ,(0.974,0.974,0.978,0.983,0.978,0.960,0.973,0.973)
                                     ,(0.990,0.990,0.992,0.994,0.991,0.983,0.989,0.989)
                                     ,(0.995,0.995,0.997,0.998,0.997,0.994,0.996,0.996)
                                     ,(1.000,1.000,1.000,1.000,1.000,1.000,1.000,1.000)
                                    ))
                  ,'model': None
                 }
                }
        return TABLE



# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

import numpy as np
from scipy.integrate import quad

class ComputeTraceNWavelength:
    '''
    ComputeTraceNWavelength computes trace and wavelength in HST/aXe definition.
    This class can be easily used following ComputeSIP, and ConfReader.
    - coef1d can be computed through ComputeSIP.
    - xyref = (xref,yref) can be computed through ComputeXYREF. Assuming the trace runs along x.
    - xgbound = (xleft,xright) specifying the computing grid which is np.arange(xleft, xright).
    Use method compute() to compute trace and wavelength, which is access ible through self.output
    - output['dydx'],output['dldp'] are coefficients for calculating trace and wavelength respectively.
    - output['xh'] is the computing grid, which is related to grism frame by xg = xref + xh.
    - output['yh'] is the trace, which is related to grism frame by yg = yref + yh.
    - output['ww'] is wavelength in A.
    '''
    def __init__(self,coef1d,xyref,xgbound):
        self.data = {'coef1d':coef1d,
                     'xyref':xyref,
                     'xgbound':xgbound
                    }
        self.output = {'xyref':xyref,
                       'xgbound':xgbound,
                       'dydx':None,
                       'dldp':None,
                       'xh':None,
                       'yh':None,
                       'ww':None
                      }
    ##########
    ##########
    ##########
    def compute(self):
        coef1d = self.data['coef1d']
        xyref = self.data['xyref']
        xgbound = self.data['xgbound']
        dydx = self._get_coef('DYDX')
        dldp = self._get_coef('DLDP')
        xh = np.arange(xgbound[0],xgbound[1])
        # trace
        yh = 0.
        for i,ii in enumerate(dydx):
            yh += ii * np.power(xh,i)
        # wavelength
        varclength = np.vectorize(self._arclength)
        arc,earc = np.array(varclength(xh,*dydx))
        ww = 0.
        for i,ii in enumerate(dldp):
            ww += ii * np.power(arc,i)
        # output
        self.output['dydx'] = dydx.copy()
        self.output['dldp'] = dldp.copy()
        self.output['xh'] = xh.copy()
        self.output['yh'] = yh.copy()
        self.output['ww'] = ww.copy()
    def _get_coef(self,key):
        coef1d = self.data['coef1d']
        tmp = {}
        for i in coef1d:
            if i.split('_')[0] == key:
                tmp[int(i.split('_')[2])] = coef1d[i]
        n = len(tmp)
        dydx = []
        for i in range(n):
            dydx.append(tmp[i])
        return dydx
    def _arclength_integrand(self,Fa,*coef):
        # compute np.sqrt(1 + np.power(dydx,2))
        # dy = sum(i=0,i=n,DYDX_A_i*np.power(dx,i))
        # dydx = (dy/dx) = sum(i=0,i=n,i*DYDX_A_i*np.power(dx,i-1))
        # Fa = dx
        # coef = [DYDX_A_0,DYDX_A_1,...]
        s = 0
        for i,ii in enumerate(coef):
            if i==0:
                continue
            s += i * ii * np.power(Fa,i-1)
        return np.sqrt(1. + np.power(s,2))
    def _arclength(self,Fa,*coef):
        # compute integrate(from 0,to Fa,integrand np.sqrt(1 + np.power(dydx,2)))
        integral,err = quad(self._arclength_integrand, 0., Fa, args=coef)
        return integral,err 



# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

from astropy.io import fits
import pandas as pd

class HeaderSummary:
    """
    header_summary is a class reading fits headers, and extracting relevant information regarding to HSTGRISM flow.
    - fits_file = (path_to_file, sci_extension)
    - params = dictionary with keyword = extension number, and values = keyword to look for in the header. If None, a preset params will be assigned.
      - Set self.params until satisfied.
    Use self.fetch() to fetch the information from headers, which can be access by self.table.
    Use self.save(Container) to save self.table to ./savefolder/saveprefix_hsum.csv given Container class object.
    For mutiple fits, use x = HeaderSummary(fits_file=None).combine(header_summary_list) where
    - header_summary_list = [x1,x2,x3,...] for each xi as HeaderSummary object.
    - This will return a dictionary of all HeaderSummary object tables. Use pandas.DataFrame to make it more readable.
    Use save(Container) method to output as ./savefolder/saveprefix_hsum.csv given Container object.
    - For multiple fits, use HeaderSummary(fits_file=None).save(Container,header_dict) where 
      - header_dict = HeaderSummary(fits_file=None).combine(header_summary_list)
    """
    def __init__(self,fits_file,params=None):
        self.fits_file = fits_file
        self.params = self._params() if params is None else params
        self.table = None
    ##################################################
    ##################################################
    ##################################################
    def save(self,container=None,header_dict=None):
        if container is None:
            raise ValueError('container must be specified to save.')
        folder,prefix = container.data['savefolder'],container.data['saveprefix']
        string = './{0}/{1}_hsum.csv'.format(folder,prefix)
        if self.fits_file is not None:
            pd.DataFrame(self.table,index=[0]).to_csv(string)
        else:
            pd.DataFrame(header_dict).to_csv(string)
        print('Save {0}'.format(string))
    ##################################################
    ##################################################
    ##################################################
    def combine(self,header_summary_list):
        """
        This function combines all HeaderSummary objects in the header_summary_list to a single object.
        To use, call HeaderSummary(fits_file=None).combine(header_summary_list).
        - header_summary_list = [x1,x2,x3,...] for each x as HeaderSummary object.
        It returns a ditionary combined
        """
        output_header = {}
        output_header['index'] = []
        output_header['filepath'] = []
        output_header['sci_ext_num'] = []
        for ii,i in enumerate(header_summary_list):
            index = ii
            header = header_summary_list[ii].table
            output_header['index'].append(index)
            output_header['filepath'].append(header_summary_list[ii].fits_file[0])
            output_header['sci_ext_num'].append(header_summary_list[ii].fits_file[1])
            for j in header:
                if j not in output_header.keys():
                    output_header[j] = []
                output_header[j].append(header[j])
        return output_header
    ##################################################
    ##################################################
    ##################################################
    def fetch(self):
        """
        Output a table given self.params
        """
        t = fits.open(self.fits_file[0])
        out = {}
        for i in self.params:
            try:
                h = t[i].header
            except:
                print('Cannot find extension {0} in file {1}'.format(i,self.fits_file[0]))
                continue
            for j in self.params[i]:
                try:
                    out[j] = h[j]
                except:
                    print('Cannot find keyword {0} in extension {1}.'.format(j,i))
                    out[j] = None
        self.table = out
    ##################################################
    ##################################################
    ##################################################
    def _params(self):
        """
        Output a dictionary as keyword = extension number, and values = keyword to look for in header.
        This function gives a preset.
        hdu = (extension_number,key1,key2,...)
        - sci extension number is given when instantiate.
        """
        try:
            phdu = (0,'ROOTNAME','DATE-OBS','FILTER','PA_V3','SUBARRAY','POSTARG1','POSTARG2','EXPSTART','EXPTIME',)
            shdu = (self.fits_file[1],'IDCSCALE','ORIENTAT')
            params = {phdu[0]:phdu[1:],shdu[0]:shdu[1:]}
            return params  
        except:
            return None
    


# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

from hstgrism.grismapcorr import GrismApCorr
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

class Contamination:
    """
    Contamination is a class handling computation for grism contamination.
    - Object = object of interest
    - Contaminate = another object in the FoV of Object and contaminating it
    - trace_object_csv = trace.csv of Object (generated by HSTGRISM flow)
    - trace_contaminate_csv = trace.csv of Contaminate
    - halfdy_object = a 1D integer array parallel to Object trace for extraction aperture halfdy width of Object (excluding trace; that is total width 2+halfdy_object + 1).
     - if halfdy_object = None, contamination region would not be computed.
    - instrument = string for available instrument list in hstgrism.grismapcorr.GrismApCorr.
     - if not None, self.compute() will grab aperture correction values (apcorr) from GrismApCorr table given instrument.
     - self.apcorr to access apcorr as a dict with key as halfdy in pixel unit and value as 1D array parallel to self.combine_df.ww_contaminate
       - Note: for ww_contaminate where it is None, we replaced it with ww_contaminate.min(). This would not cause any conflict at the end since we will place None back.
     - self.combine_df after running self.compute() will also have columns corresponding to apcorr_contaminate_in and apcorr_contaminate_out.
    Use self.compute() to start computation after properly specified information.
    - if halfdy_object is not None, contamination region would be computed.
    Use self.save(container) to save outputs to ./savefolder/saveprefix_savesuffix.extension
    - savefolder, saveprefix is controlled by Container class (see hstgrism.container.Container)
    - savesuffix.extension is preset:
     - comtaminateTrace.csv = merge trace of Contaminate to Object trace frame, accessible by self.combine_df
    Use self.show(save,container) to show traces, extraction region of Object, and contamination region
    - if save=True, the plot is saved at ./savefolder/saveprefix_contamination.plotformat where savefolder, saveprefix, and plotformat are defined by container (see hstgrism.container.Container)
    """
    def __init__(self,trace_object_csv,trace_contaminate_csv,halfdy_object=None,instrument=None):
        self.trace_object = pd.read_csv(trace_object_csv)
        self.trace_contaminate = pd.read_csv(trace_contaminate_csv)
        self.halfdy_object = halfdy_object
        self.instrument = instrument
    def compute(self):
        xg_object = np.ceil(self.trace_object.xh.values + self.trace_object.xyref[0]).astype(int)
        yg_object = np.ceil(self.trace_object.yh.values + self.trace_object.xyref[1]).astype(int)
        ww_object = self.trace_object.ww.values
        xg_contaminate = np.ceil(self.trace_contaminate.xh.values + self.trace_contaminate.xyref[0]).astype(int)
        yg_contaminate = np.ceil(self.trace_contaminate.yh.values + self.trace_contaminate.xyref[1]).astype(int)
        ww_contaminate = self.trace_contaminate.ww.values
        yg_contaminate_wrt_object = np.array([],dtype=object)
        ww_contaminate_wrt_object = np.array([],dtype=object)
        for ii,i in enumerate(xg_object):
            if i in xg_contaminate:
                m = np.argwhere(xg_contaminate == i).flatten()
                yg_contaminate_wrt_object = np.concatenate((yg_contaminate_wrt_object,yg_contaminate[m]))
                ww_contaminate_wrt_object = np.concatenate((ww_contaminate_wrt_object,ww_contaminate[m]))
            else:
                yg_contaminate_wrt_object = np.concatenate((yg_contaminate_wrt_object,[None]))
                ww_contaminate_wrt_object = np.concatenate((ww_contaminate_wrt_object,[None]))
        combine_df = {'xg_object':xg_object,'yg_object':yg_object,'ww_object':ww_object,'yg_contaminate':yg_contaminate_wrt_object,'ww_contaminate':ww_contaminate_wrt_object}
        combine_df = pd.DataFrame(combine_df)
        self.combine_df = combine_df  
        ##### make halfdy_contaminate_in and out #####
        if self.halfdy_object is not None:
            tx = self.combine_df.xg_object.values.copy()
            ty = self.combine_df.yg_object.values.copy()
            tyc = self.combine_df.yg_contaminate.copy()
            halfdy_contaminate_in = np.full_like(tx,None,dtype=object)
            halfdy_contaminate_out = np.full_like(tx,None,dtype=object)
            for ii,i in enumerate(tx):
                if tyc[ii] is not None:
                    halfdy_contaminate_in[ii] = np.abs(tyc[ii] - ty[ii]) - self.halfdy_object[ii] - 1
                    halfdy_contaminate_out[ii] = np.abs(tyc[ii] - ty[ii]) + self.halfdy_object[ii]
            self.combine_df['halfdy_contaminate_in'] = halfdy_contaminate_in.copy()
            self.combine_df['halfdy_contaminate_out'] = halfdy_contaminate_out.copy()  
        ##### prepare apcorr #####
        if self.instrument is not None:
            halfdy_list = np.concatenate((self.combine_df.halfdy_contaminate_in.unique(),self.combine_df.halfdy_contaminate_out.unique()))
            apcorr = {}
            for i in halfdy_list:
                if i is None:
                    continue
                halfdy_contaminate = i
                apsizepix = halfdy_contaminate*2 + 1
                apsizepix = np.full_like(self.combine_df.ww_contaminate.values,apsizepix,dtype=float)
                apsizearcsec = apsizepix * GrismApCorr().table[self.instrument]['scale']
                tww = self.combine_df.ww_contaminate.values.astype(float).copy()
                tww[np.isnan(tww)] = tww[np.isfinite(tww)].min()
                apcorrobj = GrismApCorr(instrument=self.instrument,apsize=apsizearcsec,wave=tww,aptype='diameter',apunit='arcsec',waveunit='A')
                apcorrobj.compute()
                apcorr[i] = apcorrobj.data['apcorr'] 
            self.apcorr = apcorr
            ##### prepare apcorr_contaminate_in and apcorr_contaminate_out #####
            ##### place them into combine_df #####
            tx = self.combine_df.halfdy_contaminate_in.values.copy()
            apcorr_contaminate_in = np.array([],dtype=object)
            for ii,i in enumerate(tx):
                if i is None:
                    apcorr_contaminate_in = np.concatenate((apcorr_contaminate_in,np.array([None])))
                else:
                    apcorr_contaminate_in = np.concatenate((apcorr_contaminate_in,np.array([self.apcorr[i][ii]])))
            tx = combine_df.halfdy_contaminate_out.values.copy()
            apcorr_contaminate_out = np.array([],dtype=object)
            for ii,i in enumerate(tx):
                if i is None:
                    apcorr_contaminate_out = np.concatenate((apcorr_contaminate_out,np.array([None])))
                else:
                    apcorr_contaminate_out = np.concatenate((apcorr_contaminate_out,np.array([self.apcorr[i][ii]])))
            self.combine_df['apcorr_contaminate_in'] = apcorr_contaminate_in
            self.combine_df['apcorr_contaminate_out'] = apcorr_contaminate_out
    def save(self,container=None):
        if container is None:
            raise ValueError('container must be specified to save.')
        string = './{0}/{1}_contaminateTrace.csv'.format(container.data['savefolder'],container.data['saveprefix'])
        self.combine_df.to_csv(string)
        print('Save {0}'.format(string))
    ##########
    ##########
    ##########
    def show(self,figsize=(10,10),
             object_color='black',object_ls=':',object_marker='x',
             contaminate_color='red',contaminate_ls=':',contaminate_marker='x',
             fontsize=12,
             save=False,container=None,
            ):
        plt.figure(figsize=figsize)
        combine_df = self.combine_df
        halfdy_object = self.halfdy_object
        plt.plot(combine_df.xg_object,combine_df.yg_object,ls=object_ls,color=object_color,label='Object trace')
        plt.plot(combine_df.xg_object,combine_df.yg_object + halfdy_object,color=object_color,marker=object_marker,label='Object aperture (inclusive)')
        plt.plot(combine_df.xg_object,combine_df.yg_object - halfdy_object,color=object_color,marker=object_marker)
        plt.plot(combine_df.xg_object,combine_df.yg_contaminate,color=contaminate_color,ls=contaminate_ls,label='Contaminate trace')
        plt.plot(combine_df.xg_object,combine_df.yg_contaminate - combine_df.halfdy_contaminate_in,color=contaminate_color,marker=contaminate_marker,label='Contaminate aperture (inclusive)')
        plt.plot(combine_df.xg_object,combine_df.yg_contaminate - combine_df.halfdy_contaminate_out,color=contaminate_color,marker=contaminate_marker)
        plt.xlabel('pixX',fontsize=fontsize)
        plt.ylabel('pixY',fontsize=fontsize)
        plt.title('Contamination',fontsize=fontsize)
        plt.legend(loc=(1.01,0.))
        plt.tight_layout()
        if save:
            if container is None:
                raise ValueError('container must be specified to save')
            string = './{0}/{1}_contamination.{2}'.format(container.data['savefolder'],container.data['saveprefix'],container.data['plotformat'])
            plt.savefig(string,plotformat=container.data['plotformat'],bbox_inches='tight')
            print('Save {0}'.format(string))          
            


# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

from astropy.io import fits
import copy
import numpy as np
import pandas as pd

class FlatField:
    '''
    FlatField computes a flatfield image in aXe definition.
    This requires flatfile produced for aXe, and trace.csv produced by hstgrism package.
    - container = Container object
    - xyref and dldp can be found in trace.csv
    Use method compute() to generate flatfield image, which is accessible through self.flatfield.
    Use method save() to save the flatfield image in full frame as ./savefolder/saveprefix_fullflate.fits
    - if save(do_cutout=True), this will also save a cutout, given bbx0x,bb1x,bb0y,bb1y (which can be found in bbcorner.csv if produced by earlier step).
        The cutout is saved as ./savefolder/saveprefix_cutflat.fits
    '''
    def __init__(self,container,flatfile,xyref,dldp):
        self.container = container
        self.flatfile = flatfile
        self.xyref = xyref
        self.dldp = dldp
        self.data = {'WMIN':None,
                     'WMAX':None,
                     'FLAT':None,
                     'NAX1':None,
                     'NAX2':None,
                     'WW':None
                    }
        self.flatfield = None
        self._get()
    def _get(self):
        # basic info
        tmp = fits.open(self.flatfile)
        wmin,wmax = tmp[0].header['WMIN'],tmp[0].header['WMAX']
        nax1,nax2 = tmp[0].header['NAXIS1'],tmp[0].header['NAXIS2']
        self.data['WMIN'] = wmin
        self.data['WMAX'] = wmax
        self.data['NAX1'] = nax1
        self.data['NAX2'] = nax2
        # flat
        n = len(tmp)
        tmpp = {}
        for i in range(n):
            tmpp[i] = tmp[i].data.copy()
        self.data['FLAT'] = copy.deepcopy(tmpp)
        # wavelength
        dldp = self.dldp
        m = np.isfinite(dldp)
        dldp = dldp[m].values
        xg = np.arange(self.data['NAX1'])
        xref = int(self.xyref[0])
        xh = xg-xref
        ww = np.full_like(xh,0.,dtype=float)
        for i,ii in enumerate(dldp):
            ww += ii * np.power(xh,i)
        self.data['WW'] = ww
        self.data['XG'] = xg
        self.data['XREF'] = xref
    ##########
    ##########
    ##########
    def compute(self):
        # dimension test
        if self.data['WW'].shape[0] != self.data['NAX1']:
            raise ValueError('Dimension of WW != NAX1')
        # set parametric wavelength [0,1]
        wmin,wmax,ww = self.data['WMIN'],self.data['WMAX'],self.data['WW']
        paramww = (ww-wmin)/(wmax-wmin)
        paramww[paramww<0.] = 0.
        paramww[paramww>1.] = 1.
        paramww,_ = np.meshgrid(paramww,np.arange(self.data['NAX2']))
        # compute
        tmp = self.data['FLAT']
        tmpp = np.full_like(paramww,0.,dtype=float)
        for i in tmp:
            tmpp += tmp[i] * np.power(paramww,i)
        self.flatfield = tmpp.copy()
    ##########
    ##########
    ##########
    def save(self,
             do_cutout=False,bb0x=None,bb1x=None,bb0y=None,bb1y=None
            ):
        if self.container is None:
            raise ValueError('container must be specified to save')
        savefolder = self.container.data['savefolder']
        saveprefix = self.container.data['saveprefix']
        # fullflat
        phdu = fits.PrimaryHDU()
        imhdu = fits.ImageHDU()
        hdul = fits.HDUList([phdu,imhdu])
        hdul[1].data = self.flatfield
        string = './{0}/{1}_fullflat.fits'.format(savefolder,saveprefix)
        hdul.writeto(string,overwrite=True)
        print('Save {0}'.format(string))
        # cutflat
        if do_cutout:
            if bb0x is None or bb1x is None or bb0y is None or bb1y is None:
                raise ValueError('bb0x,bb1x,bb0y,bb1y must be specified to save cutout')
            phdu = fits.PrimaryHDU()
            imhdu = fits.ImageHDU()
            hdul = fits.HDUList([phdu,imhdu])
            hdul[1].data = self.flatfield[bb0y:bb1y,bb0x:bb1x]
            string = './{0}/{1}_cutflat.fits'.format(savefolder,saveprefix)
            hdul.writeto(string,overwrite=True)
            print('Save {0}'.format(string))
        


# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

from polynomial2d.polynomial2d import Polynomial2D
from astropy.io import fits
import numpy as np

class Background:
    def __init__(self,gdata,norder,mdata=None,
                 rescale='default',
                 sigclip=(False,None,None),
                 container=None
                ):
        """
        Background is a class to facilitate background estimation. The routine uses Polynomial2D package (pip install polynomial2d).
        - gdata = grism data
        - norder = polynomial order
        - mdata = mask data with True for masking object
        - rescale, sigclip = parameters from Polynomial2D to perform rescaling of values (in order to improve the performance), and sigma clipping in iterative fitting process respectively.
            rescale = 'default' will set rescaling to be performed with method [-1,1] linear transformation of x1,x2,y.
            Note: Polynomial2D also uses rescalex.rescale.Rescale class (pip install rescalex) internally for rescaling.
        At the instantiation, self.poly2d is the Polynomial2D object.
        - Use self.poly2d.fit() to perform the background fitting.
        - Use self.poly2d.model['YFIT'] to access the estimated background.
        - Use self.poly2d.model['MASKFIT'] to access the final mask after iterations (with True for masked out data).
        Use method save(do_bkg,do_mask) to save outputs, given container.
        - Set do_yfit = True for saving self.poly2d.model['YFIT'] as ./savefolder/saveprefix_cutbkg.fits.
        - Set do_mask = True for saving self.poly2d.model['MASKFIT'] as ./savefolder/saveprefix_maskfit.fits.
        """
        self.gdata = gdata
        self.norder = norder
        self.container = container
        if mdata is None:
            self.mdata = np.full_like(gdata,False,dtype=bool)
        else: self.mdata = mdata
        if rescale == 'default':
            dor = True
            methodr = {'method':'linear','minmax':(-1.,1.)}
            self.rescale = (dor,methodr,methodr,methodr)
        else:
            self.rescale = rescale
        self.sigclip = sigclip
        self.poly2d = self._poly2d()
    def _poly2d(self):
        ny,nx = self.gdata.shape
        x1 = np.arange(nx)
        x2 = np.arange(ny)
        x1v,x2v = np.meshgrid(x1,x2)
        obj = Polynomial2D(x1=x1v,x2=x2v,y=self.gdata,mask=self.mdata,norder=self.norder,rescale=self.rescale,sigclip=self.sigclip)
        return obj
    ##########
    ##########
    ##########
    def save(self,do_yfit=True,do_maskfit=True):
        if self.container is None:
            raise ValueError('container must be specified to save')
        saveprefix = self.container.data['saveprefix']
        savefolder = self.container.data['savefolder']
        if do_yfit:
            phdu = fits.PrimaryHDU()
            imhdu = fits.ImageHDU()
            hdul = fits.HDUList([phdu,imhdu])
            hdul[1].data = self.poly2d.model['YFIT'].copy()
            string = './{0}/{1}_cutbkg.fits'.format(savefolder,saveprefix)
            hdul.writeto(string,overwrite=True)
            print('Save {0}'.format(string))
        if do_maskfit:
            phdu = fits.PrimaryHDU()
            imhdu = fits.ImageHDU()
            hdul = fits.HDUList([phdu,imhdu])
            hdul[1].data = self.poly2d.model['MASKFIT'].astype(int)
            string = './{0}/{1}_maskfit.fits'.format(savefolder,saveprefix)
            hdul.writeto(string,overwrite=True)
            print('Save {0}'.format(string))



import setuptools

setuptools.setup(
    name="hstgrism",
    version="8.4.0a0",
    author="Kornpob Bhirombhakdi",
    author_email="kbhirombhakdi@stsci.edu",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/bkornpob/hstgrism",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3"
        ,"License :: OSI Approved :: MIT License"
        ,"Operating System :: OS Independent"
    ],
    python_requires='>=3.'
)



# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

from hstgrism.objectmask import ObjectMask
from astropy.io import fits
import numpy as np
import pandas as pd

class Cutout2D:
    """
    Cutout2D is a class to make a 2D cutout of an image.
    - tbox['xg'] and tbox['yg'] specifies the trace.
    - tbox['halfdyup'] and tbox['halfdylow'] specified the aperture width (up and down respectively) from the trace. Note: Total width is halfdylow + 1 + halfdyup.
    - tbox defines the trace region. Only max and min of xg and yg (together with halfdyup and halfdylow) would be used to define the trace box. 
    - However, complete arrays of xg and yg are necessary when creating an object mask. Otherwise, only min and max of xg and yg are necessary to specify the cutout region.
    - bbox specifies an outer box surrounding the trace region. 
    - Note: Total x width is padxleft + (max_xg - min_xg + 1) + padxright, and total y width is padylow + halfdylow + (max_yg - min_yg + 1) + halfdyup + padyup.
    - image = full image to be cut.
    - self.bbcorner provides the left-bottom corner (bb0x,bb0y) and right-upper corner (bb1x,bb1y) computed using the specified tbox and bbox.
    Use method compute() to make a 2D cutout, which is accessible through self.cutout.
    - Set compute(do_mask = True) if user also wants an object mask, which is accessible through self.mask.
    - self.mask = True in the trace region.
    Use method save() to save the output to files, given container.
    - save(tbox,bbox,bbcorner,cutout,mask) are switches to turn on/off what user would like to save. By default, save everything.
    """
    def __init__(self,
                 tbox = {'xg':None,
                         'yg':None,
                         'halfdyup':None,
                         'halfdylow':None
                        },
                 bbox = {'padxleft':15,
                         'padxright':15,
                         'padyup':15,
                         'padylow':15
                        },
                 image = None,
                 container = None
                ):
        if tbox['xg'] is None or tbox['yg'] is None or tbox['halfdyup'] is None or tbox['halfdylow'] is None:
            raise ValueError('tbox must be specified')
        self.tbox = tbox
        self.bbox = bbox
        self.bbcorner = self._bbcorner()
        self.image = image
        self.container = container
        self.cutout = None
        self.mask = None
    def _bbcorner(self):
        minxg,maxxg = self.tbox['xg'].min(),self.tbox['xg'].max()
        minyg,maxyg = self.tbox['yg'].min(),self.tbox['yg'].max()
        bb0x = int(minxg - self.bbox['padxleft'])
        bb1x = int(maxxg + 1 + self.bbox['padxright'])
        bb0y = int(minyg - self.tbox['halfdylow'] - self.bbox['padylow'])
        bb1y = int(maxyg + 1 + self.tbox['halfdyup'] + self.bbox['padyup'])
        return {'bb0x':bb0x,'bb1x':bb1x,'bb0y':bb0y,'bb1y':bb1y}
    ##########
    ##########
    ##########
    def compute(self,do_mask=True):
        if self.image is None:
            raise ValueError('image must be specified')
        bb0x,bb1x = self.bbcorner['bb0x'],self.bbcorner['bb1x']
        bb0y,bb1y = self.bbcorner['bb0y'],self.bbcorner['bb1y']
        self.cutout = self.image[bb0y:bb1y,bb0x:bb1x].copy()
        if do_mask:
            self.mask = self._mask()
    def _mask(self):
        bb0x,bb1x = self.bbcorner['bb0x'],self.bbcorner['bb1x']
        bb0y,bb1y = self.bbcorner['bb0y'],self.bbcorner['bb1y']
        ny,nx = self.image.shape
        tx,ty = self.tbox['xg'],self.tbox['yg']
        halfdyup,halfdylow = self.tbox['halfdyup'],self.tbox['halfdylow']
        maskobj = ObjectMask(nx,ny,tx,ty,halfdyup,halfdylow)
        maskobj.compute()
        return maskobj.mask[bb0y:bb1y,bb0x:bb1x]    
    ##########
    ##########
    ##########
    def save(self,tbox=True,bbox=True,bbcorner=True,cutout=True,mask=True):
        if self.container is None:
            raise ValueError('container must be specified')
        saveprefix = self.container.data['saveprefix']
        savefolder = self.container.data['savefolder']
        if tbox:
            string = './{0}/{1}_tbox.csv'.format(savefolder,saveprefix)
            tmp = pd.DataFrame(self.tbox)
            tmp.to_csv(string)
            print('Save {0}'.format(string))
        if bbox:
            string = './{0}/{1}_bbox.csv'.format(savefolder,saveprefix)
            tmp = pd.DataFrame.from_dict(self.bbox, orient='index').T
            tmp.to_csv(string)
            print('Save {0}'.format(string))
        if bbcorner:
            string = './{0}/{1}_bbcorner.csv'.format(savefolder,saveprefix)
            tmp = pd.DataFrame.from_dict(self.bbcorner, orient='index').T
            tmp.to_csv(string)
            print('Save {0}'.format(string))
        if cutout:
            phdu = fits.PrimaryHDU()
            imhdu = fits.ImageHDU()
            hdul = fits.HDUList([phdu,imhdu])
            hdul[1].data = self.cutout
            string = './{0}/{1}_cutout.fits'.format(savefolder,saveprefix)
            hdul.writeto(string,overwrite=True)
            print('Save {0}'.format(string))
        if mask:
            phdu = fits.PrimaryHDU()
            imhdu = fits.ImageHDU()
            hdul = fits.HDUList([phdu,imhdu])
            hdul[1].data = self.mask.astype(int)
            string = './{0}/{1}_mask.fits'.format(savefolder,saveprefix)
            hdul.writeto(string,overwrite=True)
            print('Save {0}'.format(string))
                    


# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

from astropy.io import fits
import matplotlib.pyplot as plt
import numpy as np

def show_overview(gfile,dfile,xyd,xyref,xh,yh,ww,objname,save=False,container=None,
         params={'figsize':(10,10),
                 '221':{'minmax':(5.,99.),'cmap':'viridis','s':30,'facecolor':'None','edgecolor':'red','fontsize':12,'title':'default'},
                 '222':{'minmax':(5.,99.),'cmap':'viridis','s':30,'facecolor':'None','edgecolor':'red','padxy':(50,50),'fontsize':12,'title':'default'},
                 '223':{'minmax':(5.,99.),'cmap':'viridis','s':30,'facecolor':'None','edgecolor':'red','color':'red','ls':'-','lw':4,'alpha':0.6,'fontsize':12,'title':'default'},
                 '224':{'minmax':(5.,99.),'cmap':'viridis','s':30,'facecolor':'None','edgecolor':'red','color':'red','ls':'-','lw':4,'alpha':0.6,'padxy':(50,50),'tickperx':50,'annotate_marker':'ro','annotate_color':'red','annotate_fontsize':8,'annotate_rotation':30.,'fontsize':12,'title':'default'},
                }
        ):
    """
    show_overview plots 2x2 of direct and grism images with zoom-in to the object location.
    - gfile = (path to grism file, extension)
    - dfile = (path to direct file, extension)
    - xyd = (pixX,pixY) of the object in the direct image
    - xyref = (xref,yref) of the object in the grism image
    - xh,yh,ww = trace and wavelength map
    - objname = object name which will be used for the plot title
    - save = True if user wants an output file for the plot, given container
    - params = customizable plot parameters
    """
    pixx,pixy = xyd
    xref,yref = xyref
    xg,yg = xh+xref,yh+yref
    plt.figure(figsize=params['figsize'])
    # 221
    ax1 = plt.subplot(2,2,1)
    fontsize = params['221']['fontsize']
    tmpdata = fits.open(dfile[0])[dfile[1]].data
    m = np.isfinite(tmpdata)
    vmin,vmax = np.percentile(tmpdata[m],params['221']['minmax'][0]),np.percentile(tmpdata[m],params['221']['minmax'][1])
    ax1.imshow(tmpdata,origin='lower',cmap=params['221']['cmap'],vmin=vmin,vmax=vmax)
    ax1.scatter(pixx,pixy,s=params['221']['s'],facecolor=params['221']['facecolor'],edgecolor=params['221']['edgecolor'])
    if params['221']['title'] == 'default':
        tmpheader = fits.open(dfile[0])[0].header
        string = '{0} {1} {2} {3}'.format(objname,tmpheader['ROOTNAME'],tmpheader['DATE-OBS'],tmpheader['FILTER'])
        string += '\nEXPSTART={0:.3f} EXPTIME={1:.3f}'.format(tmpheader['EXPSTART'],tmpheader['EXPTIME'])
    else:
        string = params['221']['title']
    ax1.set_title(string,fontsize=fontsize)
    # 222
    ax2 = plt.subplot(2,2,2)
    fontsize = params['222']['fontsize']
    dx,dy = params['222']['padxy']
    tmpdata = fits.open(dfile[0])[dfile[1]].data
    m = np.isfinite(tmpdata)
    vmin,vmax = np.percentile(tmpdata[m],params['222']['minmax'][0]),np.percentile(tmpdata[m],params['222']['minmax'][1])
    ax2.imshow(tmpdata,origin='lower',cmap=params['222']['cmap'],vmin=vmin,vmax=vmax)
    ax2.scatter(pixx,pixy,s=params['222']['s'],facecolor=params['222']['facecolor'],edgecolor=params['222']['edgecolor'])
    ax2.set_xlim(pixx-dx,pixx+dx)
    ax2.set_ylim(pixy-dy,pixy+dy)
    if params['222']['title'] == 'default':
        string = 'xyd = {0:.1f},{1:.1f}'.format(pixx,pixy)
    else:
        string = params['222']['title']
    ax2.set_title(string,fontsize=fontsize)
    # 223
    ax3 = plt.subplot(2,2,3)
    fontsize = params['223']['fontsize']
    tmpdata = fits.open(gfile[0])[gfile[1]].data
    m = np.isfinite(tmpdata)
    vmin,vmax = np.percentile(tmpdata[m],params['223']['minmax'][0]),np.percentile(tmpdata[m],params['223']['minmax'][1])
    ax3.imshow(tmpdata,origin='lower',cmap=params['223']['cmap'],vmin=vmin,vmax=vmax)
    ax3.plot(xg,yg,color=params['223']['color'],ls=params['223']['ls'],lw=params['223']['lw'],alpha=params['223']['alpha'])
    if params['223']['title'] == 'default':
        tmpheader = fits.open(gfile[0])[0].header
        string = '{0} {1} {2} {3}'.format(objname,tmpheader['ROOTNAME'],tmpheader['DATE-OBS'],tmpheader['FILTER'])
        string += '\nEXPSTART={0:.3f} EXPTIME={1:.3f}'.format(tmpheader['EXPSTART'],tmpheader['EXPTIME'])
    else:
        string = params['223']['title']   
    ax3.set_title(string,fontsize=fontsize)
    # 224
    ax4 = plt.subplot(2,2,4)
    fontsize = params['224']['fontsize']
    tickperx = params['224']['tickperx']
    dx,dy = params['224']['padxy']
    annotate_marker = params['224']['annotate_marker']
    annotate_color = params['224']['annotate_color']
    annotate_fontsize = params['224']['annotate_fontsize']
    annotate_rotation = params['224']['annotate_rotation']
    tmpdata = fits.open(gfile[0])[gfile[1]].data
    m = np.isfinite(tmpdata)    
    vmin,vmax = np.percentile(tmpdata[m],params['224']['minmax'][0]),np.percentile(tmpdata[m],params['224']['minmax'][1])
    ax4.imshow(tmpdata,origin='lower',cmap=params['224']['cmap'],vmin=vmin,vmax=vmax)
    ax4.plot(xg,yg,color=params['224']['color'],ls=params['224']['ls'],lw=params['224']['lw'],alpha=params['224']['alpha'])
    for i,ii in enumerate(xg):
        if (i in {0,len(xg)-1}) or (np.mod(i,tickperx)==0):
            label = '{0}A'.format(int(ww[i]))
            ax4.plot(xg[i],yg[i],annotate_marker)
            ax4.annotate(label,(xg[i],yg[i]),
                         textcoords='offset points',
                         xytext=(0,10),
                         ha='center',
                         fontsize=annotate_fontsize,
                         rotation=annotate_rotation,
                         color=annotate_color
                        )
    ax4.set_xlim(xg.min()-dx,xg.max()+dx)
    ax4.set_ylim(yg.min()-dy,yg.max()+dy)  
    if params['224']['title'] == 'default':
        string = 'xyref = {0:.1f},{1:.1f}'.format(xref,yref)        
    else:
        string = params['224']['title']   
    ax4.set_title(string,fontsize=fontsize)
    plt.tight_layout()
    if save:
        if container is None:
            raise ValueEror('container must be specified to save.')
        saveprefix = container.data['saveprefix']
        savefolder = container.data['savefolder']
        saveformat = container.data['plotformat']
        string = './{2}/{0}_overview.{1}'.format(saveprefix,saveformat,savefolder)
        plt.savefig(string,format=saveformat,bbox_inches='tight')
        print('Save {0}\n'.format(string))


